@@ -79,6 +79,8 @@ PhoneticEditor::PhoneticEditor (ZhuyinProperties & props, Config & config)
         loadEasySymbolFile (path) ||
         loadEasySymbolFile (PKGDATADIR G_DIR_SEPARATOR_S "easysymbol.txt");
     g_free(path);
+
+    m_moved_left = FALSE;
 }
 
 PhoneticEditor::~PhoneticEditor (void)
@@ -109,14 +111,82 @@ PhoneticEditor::processEscape (guint keyval, guint keycode,
     if (cmshm_filter (modifiers) != 0)
         return TRUE;
 
-    reset ();
-    return TRUE;
+    if (STATE_INPUT == m_input_state) {
+        reset ();
+        update ();
+        return TRUE;
+    }
+
+    if (STATE_CANDIDATE_SHOWN == m_input_state ||
+        STATE_BUILTIN_SYMBOL_SHOWN == m_input_state ||
+        STATE_BOPOMOFO_SYMBOL_SHOWN == m_input_state ||
+        STATE_USER_SYMBOL_LIST_ALL == m_input_state ||
+        STATE_USER_SYMBOL_SHOWN == m_input_state) {
+
+        m_input_state = STATE_INPUT;
+
+        if (m_moved_left) {
+            moveCursorRight ();
+            m_moved_left = FALSE;
+        }
+
+        update ();
+        return TRUE;
+    }
+
+    return FALSE;
 }
 
+gboolean
+PhoneticEditor::processSpace (guint keyval, guint keycode,
+                              guint modifiers)
+{
+    if (IBUS_space != keyval && IBUS_KP_Space != keyval)
+        return FALSE;
+
+    if (STATE_INPUT == m_input_state) {
+        if (cmshm_filter (modifiers) != 0)
+            return FALSE;
+
+        if (m_text.empty ()) {
+            assert (is_half_english (' '));
+            if (m_props.modeFullEnglish ()) {
+                String english;
+                half_english_to_full_english (keyval, english);
+                commit (english);
+            } else {
+                String english = ' ';
+                commit (english);
+            }
+            return TRUE;
+        }
+
+        if (!m_config.spaceShowCandidates ()) {
+            return insert (keyval, keycode, modifiers);
+        }
+    }
+
+    if (STATE_CANDIDATE_SHOWN == m_input_state ||
+        STATE_BUILTIN_SYMBOL_SHOWN == m_input_state ||
+        STATE_BOPOMOFO_SYMBOL_SHOWN == m_input_state ||
+        STATE_USER_SYMBOL_LIST_ALL == m_input_state ||
+        STATE_USER_SYMBOL_SHOWN == m_input_state) {
+        if (cmshm_filter (modifiers) != 0)
+            return TRUE;
+
+        selectCandidate (m_lookup_table.cursorPos ());
+
+        updateZhuyin ();
+        update ();
+        return TRUE;
+    }
+
+    return FALSE;
+}
 
 gboolean
-PhoneticEditor::processEnter (guint keyval, guint keycode,
-                              guint modifiers)
+PhoneticEditor::processCommit (guint keyval, guint keycode,
+                               guint modifiers)
 {
     if (IBUS_Return != keyval && IBUS_KP_Enter != keyval)
         return FALSE;
@@ -132,11 +202,18 @@ PhoneticEditor::processEnter (guint keyval, guint keycode,
 }
 
 gboolean
-PhoneticEditor::processSpace (guint keyval, guint keycode,
-                              guint modifiers)
+PhoneticEditor::processSelectCandidate (guint keyval, guint keycode,
+                                        guint modifiers)
 {
-    if (IBUS_space != keyval && IBUS_KP_Space != keyval)
+
+    switch (keyval) {
+    case IBUS_Return:
+    case IBUS_KP_Enter:
+        break;
+
+    default:
         return FALSE;
+    }
 
     if (cmshm_filter (modifiers) != 0)
         return TRUE;
@@ -224,6 +301,12 @@ PhoneticEditor::processShowCandidateKey (guint keyval, guint keycode,
         switch (keyval) {
         case IBUS_Down:
         case IBUS_KP_Down:
+            if (m_config.candidatesAfterCursor () &&
+                m_cursor == get_enhanced_text_length (m_text)) {
+                moveCursorLeft ();
+                m_moved_left = TRUE;
+            }
+
             /* check phonetic or symbol section here */
             prepareCandidates ();
             break;
@@ -232,6 +315,12 @@ PhoneticEditor::processShowCandidateKey (guint keyval, guint keycode,
         case IBUS_KP_Up:
             m_lookup_table.clear ();
             m_input_state = STATE_INPUT;
+
+            if (m_moved_left) {
+                moveCursorRight ();
+                m_moved_left = FALSE;
+            }
+
             break;
 
         default:
@@ -490,8 +579,10 @@ PhoneticEditor::selectCandidate (guint index)
     if (STATE_CANDIDATE_SHOWN == m_input_state) {
         int offset = m_phonetic_section->selectCandidate (index);
 
+#if 0
         if (0 == offset)
             return FALSE;
+#endif
 
         m_cursor += offset;
         m_input_state = STATE_INPUT;
@@ -510,6 +601,9 @@ PhoneticEditor::selectCandidate (guint index)
         if (0 == offset)
             return FALSE;
 
+        if (!m_config.candidatesAfterCursor ())
+            m_cursor --;
+
         erase_input_sequence (m_text, m_cursor, 1);
         insert_symbol (m_text, m_cursor, symbols->getType (),
                        symbols->getLookup (), symbols->getChoice ());
@@ -577,42 +671,9 @@ PhoneticEditor::selectCandidateInPage (guint index)
     return selectCandidate (index);
 }
 
-gboolean
-PhoneticEditor::removeCharBefore (void)
+guint
+PhoneticEditor::getCursorLeft (void)
 {
-    if (G_UNLIKELY (m_cursor == 0))
-        return FALSE;
-
-    m_cursor --;
-    erase_input_sequence (m_text, m_cursor, 1);
-
-    updateZhuyin ();
-    update ();
-
-    return TRUE;
-}
-
-gboolean
-PhoneticEditor::removeCharAfter (void)
-{
-    if (G_UNLIKELY (m_cursor ==
-                    get_enhanced_text_length (m_text)))
-        return FALSE;
-
-    erase_input_sequence (m_text, m_cursor, 1);
-
-    updateZhuyin ();
-    update ();
-
-    return TRUE;
-}
-
-gboolean
-PhoneticEditor::moveCursorLeft (void)
-{
-    if (G_UNLIKELY (m_cursor == 0))
-        return FALSE;
-
     /* NOTE: adjust cursor when in parsed phonetic section. */
     const String & enhanced_text = m_text;
     guint cursor = m_cursor;
@@ -653,24 +714,16 @@ PhoneticEditor::moveCursorLeft (void)
 
             /* align to the begin of chewing key. */
             /* restore cursor variable. */
-            m_cursor = m_cursor - (cursor + 1 - begin);
-            update ();
-            return TRUE;
+            return m_cursor - (cursor + 1 - begin);
         }
     }
 
-    m_cursor --;
-    update ();
-    return TRUE;
+    return m_cursor - 1;
 }
 
-gboolean
-PhoneticEditor::moveCursorRight (void)
+guint
+PhoneticEditor::getCursorRight (void)
 {
-    if (G_UNLIKELY (m_cursor ==
-                    get_enhanced_text_length (m_text)))
-        return FALSE;
-
     /* NOTE: adjust cursor when in parsed phonetic section. */
     const String & enhanced_text = m_text;
     guint cursor = m_cursor;
@@ -683,7 +736,7 @@ PhoneticEditor::moveCursorRight (void)
 
     section_t type = probe_section_quick (enhanced_text, start_pos);
 
-    /* only when in phonetic section, need adjustments.  */
+    /* when in phonetic section, need adjustments.  */
     if (PHONETIC_SECTION == type) {
         String section;
         get_phonetic_section (enhanced_text, start_pos, end_pos, section);
@@ -695,7 +748,7 @@ PhoneticEditor::moveCursorRight (void)
         assert (cursor < section_len);
         assert (parsed_len <= section_len);
 
-        /* only when in parsed phonetic section, need adjustments.  */
+        /* when in parsed phonetic section, need adjustments.  */
         if (cursor < parsed_len) {
             guint16 offset = 0;
             zhuyin_get_zhuyin_key_rest_offset (instance, cursor, &offset);
@@ -714,20 +767,71 @@ PhoneticEditor::moveCursorRight (void)
                 (instance, key_rest, &begin, NULL);
 
                 /* align to the begin of chewing key. */
-                m_cursor = m_cursor + (begin - cursor);
-                update ();
-                return TRUE;
+                return m_cursor + (begin - cursor);
             } else {
                 assert (offset == len);
                 /* align to the end of parsed phonetic section. */
-                m_cursor = m_cursor + (parsed_len - cursor);
-                update ();
-                return TRUE;
+                return m_cursor + (parsed_len - cursor);
             }
         }
     }
 
-    m_cursor ++;
+    /* for symbol section. */
+    return m_cursor + 1;
+}
+
+
+gboolean
+PhoneticEditor::removeCharBefore (void)
+{
+    if (G_UNLIKELY (m_cursor == 0))
+        return FALSE;
+
+    guint cursor = getCursorLeft ();
+    erase_input_sequence (m_text, cursor, m_cursor - cursor);
+    m_cursor = cursor;
+
+    updateZhuyin ();
+    update ();
+
+    return TRUE;
+}
+
+gboolean
+PhoneticEditor::removeCharAfter (void)
+{
+    if (G_UNLIKELY (m_cursor ==
+                    get_enhanced_text_length (m_text)))
+        return FALSE;
+
+    guint cursor = getCursorRight ();
+    erase_input_sequence (m_text, m_cursor, cursor - m_cursor);
+
+    updateZhuyin ();
+    update ();
+
+    return TRUE;
+}
+
+gboolean
+PhoneticEditor::moveCursorLeft (void)
+{
+    if (G_UNLIKELY (m_cursor == 0))
+        return FALSE;
+
+    m_cursor = getCursorLeft ();
+    update ();
+    return TRUE;
+}
+
+gboolean
+PhoneticEditor::moveCursorRight (void)
+{
+    if (G_UNLIKELY (m_cursor ==
+                    get_enhanced_text_length (m_text)))
+        return FALSE;
+
+    m_cursor = getCursorRight ();
     update ();
     return TRUE;
 }
@@ -955,11 +1059,26 @@ PhoneticEditor::prepareCandidates (void)
     size_t index = 0;
     size_t start_pos = 0, end_pos = 0;
 
-    probe_section_start (enhanced_text, m_cursor,
-                         cursor, index, start_pos);
+    /* hack here: we use "m_cursor - 1" to determine the section type
+       for candidates before cursor. */
+    if (m_config.candidatesAfterCursor ())
+        probe_section_start (enhanced_text, m_cursor,
+                             cursor, index, start_pos);
+    else
+        probe_section_start (enhanced_text, m_cursor - 1,
+                             cursor, index, start_pos);
+
+    gboolean show_candidates = FALSE;
+    if (m_config.candidatesAfterCursor ()) {
+        show_candidates = m_cursor >= 0 &&
+            m_cursor < get_enhanced_text_length (enhanced_text);
+    } else {
+        show_candidates = m_cursor > 0 &&
+            m_cursor <= get_enhanced_text_length (enhanced_text);
+    }
 
     /* deal with candidates */
-    if (m_cursor < get_enhanced_text_length (enhanced_text)) {
+    if (show_candidates) {
         section_t type = probe_section_quick (enhanced_text, start_pos);
 
         if (PHONETIC_SECTION == type) {
@@ -984,7 +1103,11 @@ PhoneticEditor::prepareCandidates (void)
                 return TRUE;
             } else {
                 m_input_state = STATE_CANDIDATE_SHOWN;
-                m_phonetic_section->initCandidates (instance, cursor);
+                if (m_config.candidatesAfterCursor ())
+                    m_phonetic_section->initCandidates (instance, cursor);
+                else
+                    /* hack here: restore cursor. */
+                    m_phonetic_section->initCandidates (instance, cursor + 1);
 
                 update ();
                 return TRUE;
