@@ -63,6 +63,8 @@ PinyinEditor::commit (void)
 void
 PinyinEditor::reset (void)
 {
+    m_preedit_text = "";
+
     PhoneticEditor::reset ();
 }
 
@@ -70,6 +72,8 @@ void
 PinyinEditor::updateZhuyin (void)
 {
     const String & enhanced_text = m_text;
+    String new_text;
+    size_t append_offset = 0;
 
     resizeInstances ();
 
@@ -84,9 +88,25 @@ PinyinEditor::updateZhuyin (void)
             get_phonetic_section (enhanced_text, start_pos, end_pos, section);
 
             zhuyin_instance_t * instance = m_instances[index];
-            zhuyin_parse_more_full_pinyins (instance, section.c_str ());
+            size_t len = zhuyin_parse_more_full_pinyins
+                (instance, section.c_str ());
+            bool valid = zhuyin_valid_zhuyin_keys (instance);
+            if (!valid) {
+                size_t valid_len = zhuyin_get_parsed_input_length (instance);
+
+                const gchar * str = NULL;
+                zhuyin_get_raw_user_input (instance, &str);
+                section = str;
+
+                /* hack here: usually only one zhuyin key will have problem,
+                   so it is safe to just move the cursor. */
+                m_cursor -= len - valid_len;
+            }
             zhuyin_guess_sentence (instance);
 
+            new_text += section;
+            append_offset += section.length ();
+
             ++index;
         }
 
@@ -95,11 +115,14 @@ PinyinEditor::updateZhuyin (void)
             get_symbol_section (enhanced_text, start_pos, end_pos,
                                 type, lookup, choice);
 
+            insert_symbol (new_text, append_offset, type, lookup, choice);
+            append_offset ++;
         }
 
         start_pos = end_pos;
     }
 
+    m_text = new_text;
     return;
 }
 
@@ -173,29 +196,35 @@ PinyinEditor::updatePreeditText (void)
 }
 
 gboolean
-PinyinEditor::insert (gint ch)
+PinyinEditor::insert (guint keyval, guint keycode, guint modifiers)
 {
-    if (IS_PINYIN (ch)) {
-        insert_phonetic (m_text, m_cursor++, ch);
+    /* let client applications to handle shortcut key event */
+    modifiers = cmshm_filter (modifiers);
+
+    if (modifiers != 0 && m_text.empty ())
+        return FALSE;
+
+    if (IS_PINYIN (keyval)) {
+        insert_phonetic (m_text, m_cursor++, keyval);
 
         updateZhuyin ();
         update ();
         return TRUE;
     }
 
-    if (insertPunct (ch)) {
+    if (insertPunct (keyval)) {
         updateZhuyin ();
         update ();
         return TRUE;
     }
 
-    if (insertEnglish (ch)) {
+    if (insertEnglish (keyval)) {
         updateZhuyin ();
         update ();
         return TRUE;
     }
 
-    if (insertNumbers (ch)) {
+    if (insertNumbers (keyval)) {
         updateZhuyin ();
         update ();
         return TRUE;
@@ -204,6 +233,30 @@ PinyinEditor::insert (gint ch)
     return FALSE;
 }
 
+gboolean
+PinyinEditor::processSpace (guint keyval, guint keycode, guint modifiers)
+{
+    if (IBUS_space != keyval && IBUS_KP_Space != keyval)
+        return FALSE;
+
+    if (PhoneticEditor::processSpace (keyval, keycode, modifiers))
+        return TRUE;
+
+    if (STATE_INPUT == m_input_state) {
+        if (cmshm_filter (modifiers) != 0)
+            return FALSE;
+
+        if (m_config.spaceShowCandidates ()) {
+            /* use space to show candidates. */
+            prepareCandidates ();
+            update ();
+            return TRUE;
+        }
+    }
+
+    return FALSE;
+}
+
 gboolean
 PinyinEditor::processKeyEvent (guint keyval, guint keycode, guint modifiers)
 {
@@ -218,6 +271,9 @@ PinyinEditor::processKeyEvent (guint keyval, guint keycode, guint modifiers)
     if (processEscape (keyval, keycode, modifiers))
         return TRUE;
 
+    if (processSpace (keyval, keycode, modifiers))
+        return TRUE;
+
     if (STATE_INPUT == m_input_state) {
         if (processEasySymbolKey (keyval, keycode, modifiers))
             return TRUE;
@@ -225,10 +281,10 @@ PinyinEditor::processKeyEvent (guint keyval, guint keycode, guint modifiers)
         if (processUserSymbolKey (keyval, keycode, modifiers))
             return TRUE;
 
-        if (insert (keyval))
+        if (insert (keyval, keycode, modifiers))
             return TRUE;
 
-        if (processEnter (keyval, keycode, modifiers))
+        if (processCommit (keyval, keycode, modifiers))
             return TRUE;
 
         if (processFunctionKey (keyval, keycode, modifiers))
@@ -244,7 +300,7 @@ PinyinEditor::processKeyEvent (guint keyval, guint keycode, guint modifiers)
         STATE_USER_SYMBOL_LIST_ALL == m_input_state ||
         STATE_USER_SYMBOL_SHOWN == m_input_state) {
 
-        if (processSpace (keyval, keycode, modifiers))
+        if (processSelectCandidate (keyval, keycode, modifiers))
             return TRUE;
 
         if (processCandidateKey (keyval, keycode, modifiers))
