@@ -26,6 +26,7 @@
 #include "ZYZhuyinProperties.h"
 #include "ZYTradSimpConverter.h"
 #include "ZYEnhancedText.h"
+#include "ZYSymbols.h"
 
 
 using namespace ZY;
@@ -61,13 +62,20 @@ ZhuyinEditor::commit (void)
 void
 ZhuyinEditor::reset (void)
 {
+    m_preedit_text = "";
+
     PhoneticEditor::reset ();
 }
 
 void
 ZhuyinEditor::updateZhuyin (void)
 {
+    static const char * tones[] = {" ", "ˊ", "ˇ", "ˋ", "˙", NULL};
+
     const String & enhanced_text = m_text;
+    const size_t text_len = get_enhanced_text_length (enhanced_text);
+    String new_text;
+    size_t append_offset = 0;
 
     resizeInstances ();
 
@@ -82,9 +90,59 @@ ZhuyinEditor::updateZhuyin (void)
             get_phonetic_section (enhanced_text, start_pos, end_pos, section);
 
             zhuyin_instance_t * instance = m_instances[index];
-            zhuyin_parse_more_chewings (instance, section.c_str ());
+            size_t len = zhuyin_parse_more_chewings
+                (instance, section.c_str ());
+            bool valid = zhuyin_valid_zhuyin_keys (instance);
+            if (!valid) {
+                size_t valid_len = zhuyin_get_parsed_input_length (instance);
+
+                const gchar * str = NULL;
+                zhuyin_get_raw_user_input (instance, &str);
+                section = str;
+
+                /* hack here: usually only one zhuyin key will have problem,
+                   so it is safe to just move the cursor. */
+                m_cursor -= len - valid_len;
+            }
             zhuyin_guess_sentence (instance);
 
+            new_text += section;
+            append_offset += section.length ();
+
+            /* check whether the last character is tone,
+               if not part of parsed chewing input,
+               turn the tone into symbol. */
+            if (m_cursor == text_len && end_pos == m_text.size ()) {
+
+                /* check tone symbol. */
+                const char tone = section[section.size () - 1];
+                gchar ** symbols = NULL;
+                zhuyin_in_chewing_keyboard (instance, tone, &symbols);
+
+                gboolean is_tone = FALSE;
+                /* TODO: use g_strv_contains in future. */
+                if (1 == g_strv_length (symbols)) {
+                    for (const char ** strs = tones; *strs != NULL; ++strs) {
+                        if (g_str_equal (*strs, symbols[0])) {
+                            is_tone = TRUE;
+                            break;
+                        }
+                    }
+                }
+
+                if (is_tone && section.size () > len) {
+                    erase_input_sequence (new_text, append_offset - 1, 1);
+                    insert_symbol (new_text, append_offset - 1,
+                                   BUILTIN_SYMBOL_TYPE,
+                                   "", symbols[0]);
+                    /* as we changed the last space character,
+                       reached the end of user input, exit the loop. */
+                    g_strfreev (symbols);
+                    break;
+                }
+                g_strfreev (symbols);
+            }
+
             ++index;
         }
 
@@ -92,11 +150,15 @@ ZhuyinEditor::updateZhuyin (void)
             String type, lookup, choice;
             get_symbol_section (enhanced_text, start_pos, end_pos,
                                 type, lookup, choice);
+
+            insert_symbol (new_text, append_offset, type, lookup, choice);
+            append_offset ++;
         }
 
         start_pos = end_pos;
     }
 
+    m_text = new_text;
     return;
 }
 
@@ -178,32 +240,38 @@ ZhuyinEditor::updatePreeditText (void)
 }
 
 gboolean
-ZhuyinEditor::insert (gint ch)
+ZhuyinEditor::insert (guint keyval, guint keycode, guint modifiers)
 {
+    /* let client applications to handle shortcut key event */
+    modifiers = cmshm_filter (modifiers);
+
+    if (modifiers != 0 && m_text.empty ())
+        return FALSE;
+
     gchar ** symbols = NULL;
-    if (zhuyin_in_chewing_keyboard (m_instance, ch, &symbols)) {
+    if (zhuyin_in_chewing_keyboard (m_instance, keyval, &symbols)) {
         g_strfreev (symbols);
 
-        insert_phonetic (m_text, m_cursor++, ch);
+        insert_phonetic (m_text, m_cursor++, keyval);
 
         updateZhuyin ();
         update ();
         return TRUE;
     }
 
-    if (insertPunct (ch)) {
+    if (insertPunct (keyval)) {
         updateZhuyin ();
         update ();
         return TRUE;
     }
 
-    if (insertEnglish (ch)) {
+    if (insertEnglish (keyval)) {
         updateZhuyin ();
         update ();
         return TRUE;
     }
 
-    if (insertNumbers (ch)) {
+    if (insertNumbers (keyval)) {
         updateZhuyin ();
         update ();
         return TRUE;
@@ -212,6 +280,56 @@ ZhuyinEditor::insert (gint ch)
     return FALSE;
 }
 
+gboolean
+ZhuyinEditor::processSpace (guint keyval, guint keycode, guint modifiers)
+{
+    if (IBUS_space != keyval && IBUS_KP_Space != keyval)
+        return FALSE;
+
+    if (PhoneticEditor::processSpace (keyval, keycode, modifiers))
+        return TRUE;
+
+    String new_text = m_text;
+    guint cursor = 0;
+
+    if (STATE_INPUT == m_input_state) {
+        if (cmshm_filter (modifiers) != 0)
+            return FALSE;
+
+        /* use space to show candidates. */
+        if (m_config.spaceShowCandidates ()) {
+            size_t index = 0;
+            size_t start_pos = 0, end_pos = 0;
+
+            /* detect whether the space key is part of zhuyin input. */
+            insert_phonetic (new_text, m_cursor, ' ');
+
+            probe_section_start (new_text, m_cursor,
+                                 cursor, index, start_pos);
+
+            section_t type = probe_section_quick (new_text, start_pos);
+
+            if (PHONETIC_SECTION == type) {
+                String section;
+                get_phonetic_section (new_text, start_pos, end_pos, section);
+
+                zhuyin_parse_more_chewings (m_instance, section.c_str ());
+                size_t parsed_len = zhuyin_get_parsed_input_length (m_instance);
+
+                /* part of the zhuyin string. */
+                if (cursor < parsed_len)
+                    return insert (keyval, keycode, modifiers);
+            }
+
+            prepareCandidates ();
+            update ();
+            return TRUE;
+        }
+    }
+
+    return FALSE;
+}
+
 gboolean
 ZhuyinEditor::processKeyEvent (guint keyval, guint keycode, guint modifiers)
 {
@@ -226,6 +344,9 @@ ZhuyinEditor::processKeyEvent (guint keyval, guint keycode, guint modifiers)
     if (processEscape (keyval, keycode, modifiers))
         return TRUE;
 
+    if (processSpace (keyval, keycode, modifiers))
+        return TRUE;
+
     if (STATE_INPUT == m_input_state) {
         if (processEasySymbolKey (keyval, keycode, modifiers))
             return TRUE;
@@ -233,10 +354,10 @@ ZhuyinEditor::processKeyEvent (guint keyval, guint keycode, guint modifiers)
         if (processUserSymbolKey (keyval, keycode, modifiers))
             return TRUE;
 
-        if (insert (keyval))
+        if (insert (keyval, keycode, modifiers))
             return TRUE;
 
-        if (processEnter (keyval, keycode, modifiers))
+        if (processCommit (keyval, keycode, modifiers))
             return TRUE;
 
         if (processFunctionKey (keyval, keycode, modifiers))
@@ -252,7 +373,7 @@ ZhuyinEditor::processKeyEvent (guint keyval, guint keycode, guint modifiers)
         STATE_USER_SYMBOL_LIST_ALL == m_input_state ||
         STATE_USER_SYMBOL_SHOWN == m_input_state) {
 
-        if (processSpace (keyval, keycode, modifiers))
+        if (processSelectCandidate (keyval, keycode, modifiers))
             return TRUE;
 
         if (processCandidateKey (keyval, keycode, modifiers))
